<?php

function paas_metric_profiler_boot() {
  global $_paas_queries, $_paas_query_count, $_paas_track, $_paas_vars, $_paas_cache;
  if (isset($_GET['q']) && preg_match('/^paas\\-testing/', $_GET['q'])) {
    return;
  }
  $_paas_queries = array();
  $_paas_cache = array();
  $_paas_query_count = 0;
  $_paas_track = TRUE;
  $_paas_vars = array(
    'boot' => timer_read('page'),
  );
}

function paas_metric_profiler_init() {
  global $_paas_vars, $user;
  if (preg_match('/^paas\\-testing/', $_GET['q'])) {
    return;
  }
  $_paas_vars['init'] = timer_read('page');
  $_paas_vars['uid'] = $user->uid;
  if (variable_get('dev_query', 0)) {
    variable_set('dev_query', 0);
  }
  if (function_exists('memory_get_peak_usage')) {
    $_paas_vars['memory_boot'] = memory_get_peak_usage(TRUE);
  }
  else {
    $_paas_vars['memory_boot'] = 0;
  }
  $data = paas_metric_profiler_check_static();
  $_paas_vars['boot_static'] = $data['static'];
}

function paas_metric_profiler_preprocess_page() {
  global $_paas_vars;
  $_paas_vars['preprocess_page'] = timer_read('page');
  if (function_exists('memory_get_peak_usage')) {
    $_paas_vars['memory_preprocess_page'] = memory_get_peak_usage(TRUE);
  }
  else {
    $_paas_vars['memory_preprocess_page'] = 0;
  }
}

function paas_metric_profiler_preprocess_node() {
  global $_paas_vars;
  $_paas_vars['preprocess_node'] = timer_read('page');
}

function paas_metric_profiler_preprocess_block() {
  global $_paas_vars;
  $_paas_vars['preprocess_block'] = timer_read('page');
}

function paas_metric_profiler_query_start() {
  timer_start('paas_profiler_query');
}

function paas_metric_profiler_query_stop($sql) {
  global $_paas_queries, $_paas_query_count, $_paas_track;
  
  if (isset($_GET['q']) && preg_match('/^paas\\-testing/', $_GET['q'])) {
    return;
  }
  
  if (!$_paas_track) {
    return;
  }
  
  $timer = timer_read('paas_profiler_query');
  
  if ($timer <= 0 || $timer > 3600000) {
    // Error in measure, ignore query.
    return;
  }
  
  $trace = array();
  $bt = debug_backtrace();
  foreach ($bt as $row) {
    $trace[] = (empty($row['class']) ? '' : $row['class'] . '::') . $row['function'];
  }
  array_shift($trace);
  array_shift($trace);
  array_shift($trace);
  
  ++$_paas_query_count;
  $_paas_queries[] = array($sql, implode(',', $trace), timer_read('page'), $timer);
}

function paas_metric_profiler_cache($action) {
  global $_paas_cache;
  $trace = array();
  $bt = debug_backtrace();
  foreach ($bt as $row) {
    $trace[] = (empty($row['class']) ? '' : $row['class'] . '::') . $row['function'];
  }
  array_shift($trace);
  array_shift($trace);
  $_paas_cache[] = array($action, implode(',', $trace), timer_read('page'));
}

function paas_metric_profiler_flush() {
  global $_paas_queries, $_paas_cache, $_paas_track;
  static $first = TRUE;
  $_paas_track = FALSE;
  if ($first) {
    $first = FALSE;
    db_query('TRUNCATE TABLE {paas_profiler_query}');
    db_query('TRUNCATE TABLE {paas_profiler_cache}');
  }
  db_query('LOCK TABLES {paas_profiler_query} WRITE');
  $sql = 'INSERT INTO {paas_profiler_query} (query, trace, time, timer) VALUES (\'%s\', \'%s\', %n, %n)';
  foreach ($_paas_queries as $query) {
    db_query($sql, $query);
  }
  db_query('UNLOCK TABLES');
  db_query('LOCK TABLES {paas_profiler_cache} WRITE');
  $sql = 'INSERT INTO {paas_profiler_cache} (action, trace, time) VALUES (\'%s\', \'%s\', %n)';
  foreach ($_paas_cache as $cache) {
    db_query($sql, $cache);
  }
  db_query('UNLOCK TABLES');
  $_paas_track = TRUE;
  $_paas_queries = array();
  $_paas_cache = array();
}

function paas_metric_profiler_exit() {
  global $_paas_vars, $_paas_track;
  
  $_paas_vars['exit'] = timer_read('page');
  
  // Add memory usage.
  if (function_exists('memory_get_peak_usage')) {
    $_paas_vars['memory'] = memory_get_peak_usage(TRUE);
  }
  else {
    $_paas_vars['memory'] = 0;
  }
  
  // Add information about static vars.
  $_paas_vars += paas_metric_profiler_check_static();
  
  // Flush queries to database.
  paas_metric_profiler_flush();
  
  // Stop tracking queries.
  $_paas_track = FALSE;
  
  // Write variables to database.
  db_query('TRUNCATE TABLE {paas_profiler_var}');
  $sql = 'INSERT INTO {paas_profiler_var} (name, value) VALUES (\'%s\', %n)';
  foreach ($_paas_vars as $name => $value) {
    db_query($sql, $name, $value);
  }
}

/**
 * Get memory usage for static variables.
 */
function paas_metric_profiler_check_static() {
  global $theme_key, $language;
  
  $data = array();
  $total = 0;
  $theme = empty($theme_key) ? variable_get('theme_default', 'garland') : $theme_key;
  $functions = array(
    'theme_get_registry' => array('_theme_build_registry', $theme, array(), 'phptemplate'),
    'locale' => array('cache_get', 'locale:' . $language->language, 'cache'),
    'conf' => array('_paas_metric_get_conf'),
    'menu_tree_page_data' => array('_paas_metric_get_menu_trees'),
    'drupal_get_filename' => array('_paas_metric_drupal_get_filename'),
    'drupal_load' => array('_paas_metric_drupal_get_filename'),
    'drupal_set_content' => array('drupal_set_content'),
    'drupal_set_breadcrumb' => array('drupal_set_breadcrumb'),
    'drupal_set_html_head' => array('drupal_set_html_head'),
    'drupal_add_feed' => array('drupal_add_feed'),
    'drupal_add_css' => array('drupal_add_css'),
    'drupal_add_js' => array('drupal_add_js'),
    'drupal_get_schema' => array('drupal_get_schema'),
    'menu_get_names' => array('menu_get_names'),
    'menu_set_active_trail' => array('menu_set_active_trail'),
    'list_themes' => array('list_themes'),
    'node_get_types' => array('_paas_metric_node_get_types'),
  );
  $total = 0;
  foreach ($functions as $group => $args) {
    $function = array_shift($args);
    if ($var = call_user_func_array($function, $args)) {
      $var = serialize($var);
      $startMemory = memory_get_usage();
      $var2 = unserialize($var);
      $usage = memory_get_usage() - $startMemory - PHP_INT_SIZE;
      $total += $usage;
      $usage = round($usage / 1024);
      if ($usage) {
        $data["static_$group"] = $usage;
      }
      unset($var);
      unset($var2);
    }
  }
  $data['static'] = round($total / 1024);
  return $data;
}

/**
 * Helper function to get all Drupal variables.
 * 
 * @see paas_metric_profiler_check_static
 */
function _paas_metric_get_conf() {
  global $conf;
  return $conf;
}

/**
 * Helpter function to get all the menu trees for the current page.
 * 
 * @see paas_metric_profiler_check_static 
 */
function _paas_metric_get_menu_trees() {
  global $theme_key;
  $regions = system_region_list($theme_key);
  $menus = array();
  foreach ($regions as $region => $name) {
    $blocks = block_list($region);
    foreach ($blocks as $block) {
      if ($block->module == 'menu') {
        $menus[] = $block->delta;
      }
    }
  }
  $data = array();
  $menus = array_unique($menus);
  foreach ($menus as $menu) {
    $data[$menu] = menu_tree_page_data($menu);
  }
  return $data;
}

/**
 * Helper function to determine size of static cache in drupal_load().
 * 
 * @see paas_metric_profiler_check_static
 */
function _paas_metric_drupal_load() {
  $files = array();
  $sql = 'SELECT type, name FROM {system} WHERE type IN (\'theme\', \'module\')';
  $res = db_query($sql);
  while ($item = db_fetch_object($res)) {
    $files[$item->type][$item->name] = TRUE;
  }
  return $files;
}

/**
 * Helper function to determine size of static cache in drupal_get_filename().
 * 
 * @see paas_metric_profiler_check_static
 */
function _paas_metric_drupal_get_filename() {
  $files = array();
  $sql = 'SELECT type, name FROM {system} WHERE type IN (\'theme\', \'module\')';
  $res = db_query($sql);
  while ($item = db_fetch_object($res)) {
    $files[$item->type][$item->name] = drupal_get_filename($item->type, $item->name);
  }
  return $files;
}

/**
 * Helper function to determine size of static cache in node_get_type().
 * 
 * @see paas_metric_profiler_check_static
 */
function _paas_metric_node_get_types() {
  return array(
    node_get_types('types'),
    node_get_types('names'),
  );
}

function paas_metric_profiler_call($value) {
  switch ($value) {
    case 'queries':
      paas_dump_csv('SELECT query, trace, time, timer FROM {paas_profiler_query}');
      break;
    case 'cache':
      paas_dump_csv('SELECT action, trace, time FROM {paas_profiler_cache}');
      break;
    case 'vars':
      paas_dump_csv('SELECT name, value FROM {paas_profiler_var}');
      break;
  }
}

function paas_metric_profiler_schema() {
  $schema = array();
  
  $schema['paas_profiler_var'] = array(
    'fields' => array(
      'name' => array(
        'type' => 'varchar',
        'length' => 64,
        'not null' => TRUE,
      ),
      'value' => array(
        'type' => 'float',
        'unsigned' => TRUE,
        'not null' => TRUE,
      ),
    ),
    'primary key' => array('name'),
  );
  
  $schema['paas_profiler_query'] = array(
    'fields' => array(
      'qid' => array(
        'type' => 'serial',
        'unsigned' => TRUE,
        'not null' => TRUE,
      ),
      'query' => array(
        'type' => 'text',
        'not null' => TRUE,
      ),
      'trace' => array(
        'type' => 'text',
        'not null' => TRUE,
      ),
      'time' => array(
        'type' => 'float',
        'unsigned' => TRUE,
        'not null' => TRUE,
      ),
      'timer' => array(
        'type' => 'float',
        'unsigned' => TRUE,
        'not null' => TRUE,
      ),
    ),
    'primary key' => array('qid'),
  );
  
  $schema['paas_profiler_cache'] = array(
    'fields' => array(
      'cid' => array(
        'type' => 'serial',
        'unsigned' => TRUE,
        'not null' => TRUE,
      ),
      'action' => array(
        'type' => 'text',
        'not null' => TRUE,
      ),
      'trace' => array(
        'type' => 'text',
        'not null' => TRUE,
      ),
      'time' => array(
        'type' => 'float',
        'unsigned' => TRUE,
        'not null' => TRUE,
      ),
    ),
    'primary key' => array('cid'),
  );
  
  return $schema;
}
